"use strict";
const logger = require("anyhow");
class Bitecache {
    constructor() {
        this.store = {};
        this.setup = (collection, expiresIn) => {
            if (expiresIn < 1) {
                expiresIn = 1;
            }
            if (!logger.isReady) {
                logger.setup();
            }
            if (this.store[collection]) {
                clearInterval(this.store[collection].clearTimer);
                logger.info("Bitecache.setup", collection, `Expires in ${expiresIn}s`, "Collection already exists, will overwrite it");
            }
            else {
                logger.info("Bitecache.setup", collection, `Expires in ${expiresIn}s`);
            }
            const cleanup = () => {
                this.expire(collection);
            };
            const store = {
                items: {},
                expiresIn: expiresIn,
                expireTimer: setInterval(cleanup, expiresIn * 1000),
                size: 0,
                misses: 0
            };
            this.store[collection] = store;
        };
        this.set = (collection, key, value, expiresIn) => {
            try {
                const store = this.store[collection];
                if (!store) {
                    throw new Error(`Invalid collection: ${collection}`);
                }
                if (expiresIn < 0) {
                    expiresIn = store.expiresIn;
                }
                key = key.toString();
                const now = new Date().getTime();
                const expires = expiresIn ? now + expiresIn * 1000 : now + store.expiresIn * 1000;
                const item = { data: value, expires: expires };
                store.items[key] = item;
                store.size++;
            }
            catch (ex) {
                logger.error("Bitecache.set", collection, key, ex);
                throw ex;
            }
        };
        this.get = (collection, key) => {
            try {
                const store = this.store[collection];
                if (!store) {
                    throw new Error(`Invalid collection: ${collection}`);
                }
                key = key.toString();
                const now = new Date().getTime();
                const item = store.items[key];
                if (!item) {
                    store.misses++;
                    return null;
                }
                if (item.expires <= now) {
                    delete store.items[key];
                    store.size--;
                    return null;
                }
                return item.data;
            }
            catch (ex) {
                logger.error("Bitecache.get", collection, key, ex);
                throw ex;
            }
        };
        this.del = (collection, key) => {
            try {
                const store = this.store[collection];
                if (!store) {
                    throw new Error(`Invalid collection: ${collection}`);
                }
                key = key.toString();
                if (!(key in store.items)) {
                    store.misses++;
                    return false;
                }
                delete store.items[key];
                store.size--;
                return true;
            }
            catch (ex) {
                logger.error("Bitecache.del", collection, key, ex);
                throw ex;
            }
        };
        this.merge = (collection, key, dataToMerge) => {
            try {
                const store = this.store[collection];
                if (!store) {
                    throw new Error(`Invalid collection: ${collection}`);
                }
                key = key.toString();
                if (store.items[key] && typeof store.items[key].data == "object") {
                    Object.assign(store.items[key].data, dataToMerge);
                }
            }
            catch (ex) {
                logger.error("Bitecache.merge", collection, key, ex);
                throw ex;
            }
        };
        this.expire = (collection) => {
            try {
                const store = this.store[collection];
                if (!store) {
                    throw new Error(`Invalid collection: ${collection}`);
                }
                const storeItems = Object.entries(store.items);
                const now = new Date().getTime();
                let key;
                let item;
                for ([key, item] of storeItems) {
                    if (item.expires <= now) {
                        delete store.items[key];
                        store.size--;
                    }
                }
            }
            catch (ex) {
                logger.error("Bitecache.expire", collection, ex);
                throw ex;
            }
        };
        this.clear = (collection) => {
            try {
                if (collection) {
                    const store = this.store[collection];
                    if (!store) {
                        throw new Error(`Invalid collection: ${collection}`);
                    }
                    store.items = {};
                    store.size = 0;
                    store.misses = 0;
                }
                else {
                    for (let c in this.store) {
                        this.store[c].items = {};
                        this.store[c].size = 0;
                        this.store[c].misses = 0;
                    }
                }
            }
            catch (ex) {
                logger.error("Bitecache.clear", collection, ex);
                throw ex;
            }
        };
        this.stats = (collection) => {
            try {
                const store = this.store[collection];
                if (!store) {
                    throw new Error(`Invalid collection: ${collection}`);
                }
                return {
                    size: store.size,
                    memSize: this.memSizeOf(collection),
                    misses: store.misses,
                    expiresIn: store.expiresIn
                };
            }
            catch (ex) {
                logger.error("Bitecache.stats", collection, ex);
                throw ex;
            }
        };
        this.memSizeOf = (collection) => {
            try {
                const store = this.store[collection];
                if (!store) {
                    throw new Error(`Invalid collection: ${collection}`);
                }
                const objectList = [];
                let stack = [store.items];
                let bytes = 0;
                while (stack.length) {
                    let value = stack.pop();
                    if (typeof value === "boolean") {
                        bytes += 4;
                    }
                    else if (typeof value === "string") {
                        bytes += value.length * 2;
                    }
                    else if (typeof value === "number") {
                        bytes += 8;
                    }
                    else if (typeof value === "object" && objectList.indexOf(value) === -1) {
                        objectList.push(value);
                        if (Object.prototype.toString.call(value) != "[object Array]") {
                            for (let key in value) {
                                bytes += 2 * key.length;
                            }
                        }
                        for (let key in value) {
                            stack.push(value[key]);
                        }
                    }
                }
                return bytes;
            }
            catch (ex) {
                logger.error("Bitecache.expire", collection, ex);
                throw ex;
            }
        };
    }
    static get Instance() {
        return this._instance || (this._instance = new this());
    }
    get totalSize() {
        let result = 0;
        for (let collection in this.store) {
            result += this.store[collection].size;
        }
        return result;
    }
    get totalMemSize() {
        let result = 0;
        for (let collection in this.store) {
            result += this.memSizeOf(collection);
        }
        return result;
    }
    get totalMisses() {
        let result = 0;
        for (let collection in this.store) {
            result += this.store[collection].misses;
        }
        return result;
    }
}
module.exports = Bitecache.Instance;
